// src/manip.rs

use std::{collections::HashMap, fs, path::PathBuf};

use log::error;

// ENUMS

/// Specifies a limit type, whether block or silence
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Limit {
    Suspend,
    Silence,
}

// STRUCTS

/// Wraps a host's weighted limits
#[derive(Debug, Default, Eq, PartialEq)]
pub struct LimitIndices {
    pub suspend: u16,
    pub silence: u16,
}

impl LimitIndices {
    pub fn add_limit(&mut self, limit: Limit, weight: u16) -> &mut Self {
        match limit {
            Limit::Suspend => self.suspend += weight,
            Limit::Silence => self.silence += weight,
        }
        self
    }

    pub fn normalize(&self, max: (u16, u16)) -> (u16, u16) {
        let suspend = (self.suspend as f32) / (max.0 as f32) * 100f32;
        let silence = (self.silence as f32) / (max.1 as f32) * 100f32;

        (suspend as u16, silence as u16)
    }
}

impl From<(u16, u16)> for LimitIndices {
    /// Creates indices from a tuple of two `u16` weights. Useful mostly for
    /// testing.
    fn from(value: (u16, u16)) -> Self {
        Self {
            suspend: value.0,
            silence: value.1,
        }
    }
}

/// Contains a mapping of hosts to limits as well as a trust value which is
/// used to weight limits when building a merged list
#[derive(Debug, PartialEq, Eq)]
pub struct LimitList {
    pub hosts: Vec<String>,
    pub trust: u16,
}

impl LimitList {
    pub fn from_file(path: PathBuf) -> std::io::Result<Self> {
        let mut list = LimitList {
            hosts: vec![],
            trust: 100,
        };
        let contents = fs::read_to_string(path)?;

        for host in contents.lines().filter(|line| !line.is_empty()) {
            list.hosts.push(host.to_string());
        }

        Ok(list)
    }
}

/// A map of hosts (as strings) to their limit weights
#[derive(Debug, Default, Eq, PartialEq)]
pub struct MergedLimitList {
    pub hosts: HashMap<String, LimitIndices>,
    pub trusts: (u16, u16),
}

impl MergedLimitList {
    pub fn new() -> Self {
        Self {
            hosts: HashMap::new(),
            trusts: (0, 0),
        }
    }

    pub fn add_limit_list(&mut self, src: LimitList, limit_type: Limit) -> &mut Self {
        for host in src.hosts.into_iter() {
            let entry = self.hosts.entry(host).or_default();
            entry.add_limit(limit_type, src.trust);
        }

        match limit_type {
            Limit::Suspend => self.trusts.0 += src.trust,
            Limit::Silence => self.trusts.1 += src.trust,
        }

        self
    }

    pub fn export_file(
        self,
        suspend_path: &str,
        silence_path: &str,
        thresholds: (u16, u16),
    ) -> std::io::Result<()> {
        if self.hosts.is_empty() {
            error!("Nothing to export!");
            return Ok(());
        }

        let (suspend_thresh, silence_thresh) = thresholds;
        let mut suspend_contents: String = String::default();
        let mut silence_contents: String = String::default();

        for item in self.hosts.into_iter() {
            let (suspend_trust, silence_trust) = item.1.normalize(self.trusts);

            // TODO What should the behavior be when an instance is blocked or
            // muted by most instances, but neither rises above the threshold?
            if suspend_trust >= suspend_thresh {
                suspend_contents.push_str(&(item.0.clone() + "\n"));
            }

            if silence_trust >= silence_thresh {
                silence_contents.push_str(&(item.0.clone() + "\n"));
            }
        }

        fs::write(suspend_path, alphabetize_lines(suspend_contents))?;
        fs::write(silence_path, alphabetize_lines(silence_contents))?;

        Ok(())
    }
}

pub fn alphabetize_lines(string: String) -> String {
    let mut v: Vec<&str> = string.lines().collect();
    v.sort_by_key(|a| a.to_lowercase());

    v.join("\n") + "\n"
}
